/* Seaquest Event Logging
 *
// @@@ START COPYRIGHT @@@
//
// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.
//
// @@@ END COPYRIGHT @@@
 *
 *  This library is a centralized location to define event numbers for
 *  all Trafodion components.
 *  Event(Log message) number has to follow specific pattern:
 *  Level one(first digit), "1" is used represent Trafodion.
 *  Level two (the next two digits) represents component ID, such as 01 is for Monitor
 *  Level three (the rest numbers) are for event ID within the component.
 *  Level three should be exactly 6 digits. Thus, all event numbers are 9 digits.
 */

#ifndef EVL_SQLOG_EVENTNUM_H  // guard against multiple inclusions
#define EVL_SQLOG_EVENTNUM_H

/**********************************************/
/*          Monitor Component = 01            */
/**********************************************/

// The monitor pattern for the last 6 digits is as follows:
//      Level three (the first two digits) are the module. For example:
//          cluster.cxx     = 01
//          monitor.cxx     = 02
//          etc.
//      Level four (the next two digits) are unique to the method or function.
//      Level five (the last two digits) are unique to the log event instance.


/* Module: cluster.cxx = 01 */

#define MON_CLUSTER_ALLGATHER               101010101
#define MON_CLUSTER_ASSIGNPID               101010201 // TODO: Remove after CONCURRENT_REQUESTS is fully implemented
#define MON_CLUSTER_BARRIER                 101010301
#define MON_CLUSTER_CLUSTER_1               101010401
#define MON_CLUSTER_CLUSTER_2               101010402
#define MON_CLUSTER_UCLUSTER                101010501

#define MON_CLUSTER_HANDLEOTHERNODE_1       101010601
#define MON_CLUSTER_HANDLEOTHERNODE_2       101010602
#define MON_CLUSTER_HANDLEOTHERNODE_3       101010603
#define MON_CLUSTER_HANDLEOTHERNODE_4       101010604
#define MON_CLUSTER_HANDLEOTHERNODE_5       101010605
#define MON_CLUSTER_HANDLEOTHERNODE_6       101010606
#define MON_CLUSTER_HANDLEOTHERNODE_7       101010607
#define MON_CLUSTER_HANDLEOTHERNODE_8       101010608
#define MON_CLUSTER_HANDLEOTHERNODE_9       101010609
#define MON_CLUSTER_HANDLEOTHERNODE_10      101010610
#define MON_CLUSTER_HANDLEOTHERNODE_11      101010611

#define MON_CLUSTER_HANDLEMYNODE_1          101010701
#define MON_CLUSTER_HANDLEMYNODE_2          101010702
#define MON_CLUSTER_HANDLEMYNODE_3          101010703
#define MON_CLUSTER_HANDLEMYNODE_4          101010704
#define MON_CLUSTER_HANDLEMYNODE_5          101010705
#define MON_CLUSTER_HANDLEMYNODE_6          101010706

#define MON_CLUSTER_CHECKWHODIED_1          101010801
#define MON_CLUSTER_CHECKWHODIED_2          101010802

#define MON_CLUSTER_REGROUP_1               101010901
#define MON_CLUSTER_REGROUP_2               101010902
#define MON_CLUSTER_INITCLUSTER             101011001

#define MON_CLUSTER_MARKDOWN_1              101011101
#define MON_CLUSTER_MARKDOWN_2              101011102
#define MON_CLUSTER_MARKDOWN_3              101011103
#define MON_CLUSTER_MARKDOWN_4              101011104
//#define MON_CLUSTER_MARKUP                  101011201
#define MON_CLUSTER_NODE_TM_READY_1         101011301
#define MON_CLUSTER_NODE_TM_READY_2         101011302
#define MON_CLUSTER_FORCEDOWN_1             101011401
#define MON_CLUSTER_FORCEDOWN_2             101011402
#define MON_CLUSTER_CLUSTER_MANAGER         101011501

#define MON_CLUSTER_EXPEDITEDOWN_1          101011601
#define MON_CLUSTER_EXPEDITEDOWN_2          101011602
#define MON_CLUSTER_RESPONSIVE_1            101011701
#define MON_CLUSTER_RESPONSIVE_2            101011702
#define MON_CLUSTER_RESPONSIVE_3            101011703

#define MON_CLUSTER_CONNTONEWMON_1          101011801
#define MON_CLUSTER_CONNTONEWMON_2          101011802
#define MON_CLUSTER_CONNTONEWMON_10         101011810
#define MON_CLUSTER_CONNTONEWMON_11         101011811

#define MON_CLUSTER_MERGETONEWMON_1         101011901
#define MON_CLUSTER_MERGETONEWMON_2         101011902
#define MON_CLUSTER_MERGETONEWMON_3         101011903
#define MON_CLUSTER_MERGETONEWMON_10        101011910
#define MON_CLUSTER_MERGETONEWMON_11        101011911
#define MON_CLUSTER_MERGETONEWMON_12        101011912
#define MON_CLUSTER_MERGETONEWMON_13        101011913

#define MON_CLUSTER_REINTEGRATE_1           101012001
#define MON_CLUSTER_REINTEGRATE_2           101012002
#define MON_CLUSTER_REINTEGRATE_3           101012003
#define MON_CLUSTER_REINTEGRATE_10          101012010
#define MON_CLUSTER_REINTEGRATE_11          101012011
#define MON_CLUSTER_REINTEGRATE_12          101012012
#define MON_CLUSTER_REINTEGRATE_13          101012013

#define MON_CLUSTER_REINITCLUSTER_1         101012101
#define MON_CLUSTER_REINITCLUSTER_2         101012102
#define MON_CLUSTER_REINITCLUSTER_3         101012103
#define MON_CLUSTER_REINITCLUSTER_4         101012104
#define MON_CLUSTER_REINITCLUSTER_5         101012105

#define MON_CLUSTER_CHECKWHOJOINED_1        101012201

#define MON_CLUSTER_UPDDATECLUSTER_1        101012301
#define MON_CLUSTER_UPDDATECLUSTER_2        101012302
#define MON_CLUSTER_UPDTCLUSTERSTATE_1      101012401
#define MON_CLUSTER_UPDTCLUSTERSTATE_2      101012402
#define MON_CLUSTER_ALLGATHERIB_1           101012501
#define MON_CLUSTER_ALLGATHERIB_2           101012502
#define MON_CLUSTER_ALLGATHERIB_3           101012503
#define MON_CLUSTER_ALLGATHERIB_4           101012504
#define MON_CLUSTER_EXCHANGENODEDATA_1      101012601
#define MON_CLUSTER_EXCHANGETMSYNC_1        101012701
#define MON_CLUSTER_VALIDATE_STATE_1        101012801
#define MON_CLUSTER_VALIDATE_STATE_2        101012802
#define MON_CLUSTER_VALIDATE_STATE_3        101012803
#define MON_CLUSTER_DODEVICEREQ_1           101012901
#define MON_CLUSTER_INITCONFIGCLUSTER_1     101013001
#define MON_CLUSTER_INITCONFIGCLUSTER_2     101013002
#define MON_CLUSTER_INITCONFIGCLUSTER_3     101013003
#define MON_CLUSTER_INITCONFIGCLUSTER_4     101013004

#define MON_CLUSTER_SETNEWCOMM_1            101013101
#define MON_CLUSTER_SETNEWCOMM_2            101013102

#define MON_CLUSTER_SETNEWSOCK_1            101013201
#define MON_CLUSTER_SETNEWSOCK_2            101013202
#define MON_CLUSTER_SETNEWSOCK_3            101013203

#define MON_CLUSTER_ALLGATHERSOCK_1         101013301
#define MON_CLUSTER_ALLGATHERSOCK_2         101013302
#define MON_CLUSTER_ALLGATHERSOCK_3         101013303
#define MON_CLUSTER_ALLGATHERSOCK_4         101013304
#define MON_CLUSTER_ALLGATHERSOCK_5         101013305
#define MON_CLUSTER_ALLGATHERSOCK_6         101013306
#define MON_CLUSTER_ALLGATHERSOCK_7         101013307
#define MON_CLUSTER_ALLGATHERSOCK_8         101013308
#define MON_CLUSTER_EPOLLCTL_1              101013401

#define MON_CLUSTER_INITCLUSTERSOCKS_1      101013501
#define MON_CLUSTER_INITCLUSTERSOCKS_2      101013502
#define MON_CLUSTER_INITCLUSTERSOCKS_3      101013503
#define MON_CLUSTER_INITCLUSTERSOCKS_4      101013504
#define MON_CLUSTER_INITCLUSTERSOCKS_5      101013505
#define MON_CLUSTER_INITCLUSTERSOCKS_6      101013506
#define MON_CLUSTER_INITCLUSTERSOCKS_7      101013507
#define MON_CLUSTER_INITCLUSTERSOCKS_8      101013508
#define MON_CLUSTER_INITACCEPTSOCK_1        101013601
#define MON_CLUSTER_INITACCEPTSOCK_2        101013602

#define MON_CLUSTER_MKSRVSOCK_1             101013701
#define MON_CLUSTER_MKSRVSOCK_2             101013702
#define MON_CLUSTER_MKSRVSOCK_3             101013703
#define MON_CLUSTER_MKSRVSOCK_4             101013704
#define MON_CLUSTER_MKSRVSOCK_5             101013705
#define MON_CLUSTER_MKSRVSOCK_6             101013706
#define MON_CLUSTER_MKSRVSOCK_7             101013707
#define MON_CLUSTER_MKSRVSOCK_8             101013708

#define MON_CLUSTER_MKCLTSOCK_1             101013801
#define MON_CLUSTER_MKCLTSOCK_2             101013802
#define MON_CLUSTER_MKCLTSOCK_3             101013803
#define MON_CLUSTER_MKCLTSOCK_4             101013804
#define MON_CLUSTER_MKCLTSOCK_5             101013805
#define MON_CLUSTER_MKCLTSOCK_6             101013806
#define MON_CLUSTER_MKCLTSOCK_7             101013807
#define MON_CLUSTER_MKCLTSOCK_8             101013808
#define MON_CLUSTER_MKCLTSOCK_9             101013809
#define MON_CLUSTER_MKCLTSOCK_10            101013810
#define MON_CLUSTER_MKCLTSOCK_11            101013811
#define MON_CLUSTER_MKCLTSOCK_12            101013812

#define MON_CLUSTER_CONNECT_1               101013901
#define MON_CLUSTER_CONNECT_2               101013902
#define MON_CLUSTER_CONNECT_3               101013903
#define MON_CLUSTER_CONNECT_4               101013904
#define MON_CLUSTER_CONNECT_5               101013905
#define MON_CLUSTER_CONNECT_6               101013906

#define MON_CLUSTER_CONNECTTOSELF_1         101014001
#define MON_CLUSTER_CONNECTTOSELF_2         101014002
#define MON_CLUSTER_CONNECTTOSELF_3         101014003

#define MON_CLUSTER_ACCEPTSOCK_1            101014101
#define MON_CLUSTER_ACCEPTSOCK_2            101014102
#define MON_CLUSTER_ACCEPTSOCK_3            101014103

#define MON_CLUSTER_INITSERVERSOCK_1        101014201
#define MON_CLUSTER_INITSERVERSOCK_2        101014202
#define MON_CLUSTER_INITSERVERSOCK_3        101014203
#define MON_CLUSTER_INITSERVERSOCK_4        101014204
#define MON_CLUSTER_INITSERVERSOCK_5        101014205
#define MON_CLUSTER_INITSERVERSOCK_6        101014206
#define MON_CLUSTER_INITSERVERSOCK_7        101014207

#define MON_CLUSTER_SOFTNODEDOWN_1          101014301
#define MON_CLUSTER_SOFTNODEDOWN_2          101014302
#define MON_CLUSTER_SOFTNODEDOWN_3          101014303

#define MON_CLUSTER_SOFTNODEUP_1            101014401

#define MON_CLUSTER_SETKEEPALIVESOCKOPT_1   101014501
#define MON_CLUSTER_SETKEEPALIVESOCKOPT_2   101014502
#define MON_CLUSTER_SETKEEPALIVESOCKOPT_3   101014503
#define MON_CLUSTER_SETKEEPALIVESOCKOPT_4   101014504

#define MON_CLUSTER_NO_LICENSE_VERIFIERS    101014601

#define MON_CLUSTER_ALLGATHERSOCKRECONN_1   101014701

#define MON_CLUSTER_HARDNODEUP_1            101014801

#define MON_CLUSTER_ACCEPTSOCKPEER_1        101014901
#define MON_CLUSTER_ACCEPTSOCKPEER_2        101014902

#define MON_CLUSTER_CONNECTSOCKPEER_1       101015001
#define MON_CLUSTER_CONNECTSOCKPEER_2       101015002
#define MON_CLUSTER_CONNECTSOCKPEER_3       101015003

#define MON_CLUSTER_EPOLLCTLDELETE_1        101015101

#define MON_PINGSOCKPEER_1                  101015201
#define MON_PINGSOCKPEER_2                  101015202
#define MON_PINGSOCKPEER_3                  101015203
#define MON_PINGSOCKPEER_4                  101015204

#define MON_CLUSTER_ASSIGNMONITORLEADER_1   101015301
#define MON_CLUSTER_ASSIGNMONITORLEADER_2   101015302
#define MON_CLUSTER_ASSIGNMONITORLEADER_3   101015303
#define MON_CLUSTER_ASSIGNMONITORLEADER_4   101015304

#define MON_CLUSTER_CHECKIFDONE_1           101015401

#define MON_CLUSTER_HARDNODEUPNS_1          101015501
#define MON_CLUSTER_HARDNODEUPNS_2          101015502

#define MON_CLUSTER_RECEIVESOCK_1           101015601

#define MON_CLUSTER_SENDSOCK_1              101015701

/* Module: monitor.cxx = 02 */

#define MON_MONITOR_MAIN_1                  101020101
#define MON_MONITOR_MAIN_2                  101020102
#define MON_MONITOR_MAIN_3                  101020103
#define MON_MONITOR_MAIN_4                  101020104
#define MON_MONITOR_MAIN_5                  101020105
#define MON_MONITOR_MAIN_6                  101020106
#define MON_MONITOR_MAIN_7                  101020107
#define MON_MONITOR_MAIN_8                  101020108
#define MON_MONITOR_MAIN_9                  101020109
#define MON_MONITOR_MAIN_10                 101020110
#define MON_MONITOR_MAIN_11                 101020111
#define MON_MONITOR_MAIN_12                 101020112
#define MON_MONITOR_MAIN_13                 101020113
#define MON_MONITOR_MAIN_14                 101020114
#define MON_MONITOR_MAIN_15                 101020115
#define MON_MONITOR_MAIN_16                 101020116
#define MON_MONITOR_MAIN_17                 101020117
#define MON_MONITOR_MAIN_18                 101020118
#define MON_MONITOR_TMLEADER_1              101020201
#define MON_MONITOR_TMLEADER_2              101020202
#define MON_MONITOR_DEATH_HANDLER_1         101020301
#define MON_MONITOR_DEATH_HANDLER_2         101020302
#define MON_MONITOR_DEATH_HANDLER_3         101020303
#define MON_MONITOR_DEATH_HANDLER_4         101020304
#define MON_MONITOR_DEATH_HANDLER_5         101020305
#define MON_MONITOR_DEATH_HANDLER_6         101020306
#define MON_MONITOR_DEATH_HANDLER_7         101020307
#define MON_MONITOR_BEGINREQUEST            101020401
#define MON_MONITOR_STARTPROCESS_1          101020501
#define MON_MONITOR_STARTPROCESS_2          101020502
#define MON_MONITOR_STARTPROCESS_3          101020503
#define MON_MONITOR_STARTPROCESS_4          101020504
#define MON_MONITOR_STARTPROCESS_5          101020505
#define MON_MONITOR_STARTPROCESS_6          101020506
#define MON_MONITOR_STARTPROCESS_7          101020507
#define MON_MONITOR_STARTPROCESS_8          101020508
#define MON_MONITOR_STARTPROCESS_9          101020509
#define MON_MONITOR_STARTPROCESS_10         101020510
#define MON_MONITOR_STARTPROCESS_11         101020511
#define MON_MONITOR_STARTPROCESS_12         101020512
#define MON_MONITOR_STARTPROCESS_13         101020513
#define MON_MONITOR_STARTPROCESS_14         101020514
#define MON_MONITOR_STARTPROCESS_15         101020515
#define MON_MONITOR_PROCESSINFO             101020601
#define MON_MONITOR_PROCESSREQUEST_1        101020701
#define MON_MONITOR_PROCESSREQUEST_2        101020702
#define MON_MONITOR_PROCESSREQUEST_3        101020703
#define MON_MONITOR_PROCESSREQUEST_4        101020704
#define MON_MONITOR_PROCESSREQUEST_5        101020705
#define MON_MONITOR_PROCESSREQUEST_6        101020706
#define MON_MONITOR_OPENPROCESS             101020801
#define MON_MONITOR_OPENINFO                101020901
#define MON_MONITOR_NOTIFYPROCESS_1         101021001
#define MON_MONITOR_NOTIFYPROCESS_2         101021002
#define MON_MONITOR_NOTIFYPROCESS_3         101021003
#define MON_MONITOR_NODEINFO_1              101021101
#define MON_MONITOR_NODEINFO_2              101021102
#define MON_MONITOR_NODEINFO_3              101021103
#define MON_MONITOR_NODEINFO_4              101021104
#define MON_MONITOR_COMPLETEPSTARTUP_1      101021201
#define MON_MONITOR_COMPLETEPSTARTUP_2      101021202
#define MON_MONITOR_CLOSEPROCESS            101021301
#define MON_MONITOR_GETMESSAGE_1            101021401
#define MON_MONITOR_GETMESSAGE_2            101021402
#define MON_MONITOR_GETMESSAGE_3            101021403
#define MON_MONITOR_GETMESSAGE_4            101021404
#define MON_MONITOR_GETCONF_1               101021501
#define MON_MONITOR_GETCONF_2               101021502
#define MON_MONITOR_GETCONF_3               101021503
#define MON_MONITOR_EXITPROCESS             101021601
#define MON_MONITOR_ENDREQUEST_1            101021701
#define MON_MONITOR_ENDREQUEST_2            101021702
#define MON_MONITOR_INIT_ADOPTER            101021801
#define MON_MONITOR_ADOPTER_SHUTDOWN_1      101021901
#define MON_MONITOR_BLOCK_SIG_1             101022001
#define MON_MONITOR_UNBLOCK_SIG_1           101022101
#define MON_MONITOR_PNODEINFO_1             101022201
#define MON_MONITOR_PNODEINFO_2             101022202
#define MON_MONITOR_CREATEZCLIENT_1         101022301
#define MON_MONITOR_CREATEZCLIENT_2         101022302
#define MON_MONITOR_CREATEZCLIENT_3         101022303
#define MON_MONITOR_STARTZCLIENT_1          101022401

/* Module: process.cxx = 03 */

#define MON_PROCESS_PROCESS_1               101030101
#define MON_PROCESS_PROCESS_2               101030102
#define MON_PROCESS_COMPLETEPSTARTUP_1      101030201
#define MON_PROCESS_COMPLETEPSTARTUP_2      101030202
#define MON_PROCESS_EXIT_1                  101030301
#define MON_PROCESS_EXIT_2                  101030302
#define MON_PROCESS_EXIT_3                  101030303
#define MON_PROCESS_OPEN_1                  101030401
#define MON_PROCESS_RECV_1                  101030501
#define MON_PROCESS_RECV_2                  101030502
#define MON_PROCESS_SEND_1                  101030601
#define MON_PROCESS_SEND_2                  101030602
#define MON_PROCESSCONT_PROCESSCONT_1       101030701
#define MON_PROCESSCONT_PROCESSCONT_2       101030702
#define MON_PROCESSCONT_PROCESSCONT_3       101030703
#define MON_PROCESSCONT_PROCESSCONT_4       101030704
#define MON_PROCESSCONT_ATTACHPCHECK_1      101030801
#define MON_PROCESSCONT_ATTACHPCHECK_2      101030802
#define MON_PROCESSCONT_ATTACHPCHECK_3      101030803
#define MON_PROCESSCONT_ATTACHPCHECK_4      101030804
#define MON_PROCESSCONT_CLONEPROCESS_1      101030901
#define MON_PROCESSCONT_CLONEPROCESS_2      101030902
#define MON_PROCESSCONT_CLONEPROCESS_3      101030903
#define MON_PROCESSCONT_COMPLETEPSTARTUP_1  101031001
#define MON_PROCESSCONT_CREATEPROCESS_1     101031101
#define MON_PROCESSCONT_CREATEPROCESS_2     101031102
#define MON_PROCESSCONT_CREATEPROCESS_3     101031103
#define MON_PROCESSCONT_EXITPROCESS_1       101031201
#define MON_PROCESSCONT_EXITPROCESS_2       101031202
#define MON_PROCESSCONT_OPENPROCESS_1       101031301
#define MON_PROCESSCONT_OPENPROCESS_2       101031302
#define MON_PROCESS_CREATE_1                101031401
#define MON_PROCESS_CREATE_2                101031402
#define MON_PROCESS_CREATE_3                101031403
#define MON_PROCESS_CREATE_4                101031404
#define MON_PROCESS_CREATE_5                101031405
#define MON_PROCESS_CREATE_6                101031406
#define MON_PROCESS_CREATE_7                101031407
#define MON_PROCESS_CREATE_8                101031408
#define MON_PROCESS_CREATE_9                101031409
#define MON_PROCESS_CREATE_10               101031410
#define MON_PROCESS_CREATE_11               101031411
#define MON_PROCESS_CREATE_12               101031412
#define MON_PROCESS_CREATE_13               101031413
#define MON_PROCESS_SETPROCESSSTATE_1       101031501
#define MON_PROCESS_PIDHANGUPCHECK_1        101031601
#define MON_PROCESS_PIDHANGUPCHECK_2        101031602
#define MON_PROCESS_PIDHANGUPCHECK_3        101031603
#define MON_PROCESS_CHECKFDSTATE_1          101031701
#define MON_PROCESS_CHECKFDSTATE_2          101031702
#define MON_PROCESS_CHECKFDSTATE_3          101031703
#define MON_PROCESS_CHECKFDSTATE_4          101031704
#define MON_PROCESS_CHECKFDSTATE_5          101031705
#define MON_PROCESS_PICKSTDFILE_1           101031801
#define MON_PROCESS_PICKSTDFILE_2           101031802
#define MON_PROCESS_SETUPFIFO_1             101031901
#define MON_PROCESS_SETUPFIFO_2             101031902
#define MON_PROCESS_SETUPFIFO_3             101031903
#define MON_PROCESS_SETUPFIFO_4             101031904
#define MON_PROCESS_SETUPFIFO_5             101031905
#define MON_PROCESS_SETUPFIFO_6             101031906
#define MON_PROCESS_SETUPFIFO_7             101031907
#define MON_PROCESS_SETUPFIFO_8             101031908
#define MON_PROCESS_SETUPPIPE_1             101032001
#define MON_PROCESS_SETUPPIPE_2             101032002
#define MON_PROCESS_SETUPPIPE_3             101032003
#define MON_PROCESS_SETUPPIPE_4             101032004
#define MON_PROCESS_CLOSE_1                 101032101
#define MON_PROCESS_CLOSE_2                 101032102
#define MON_PROCESS_PERSIST_1               101032201
#define MON_PROCESS_PERSIST_2               101032202
#define MON_PROCESS_PERSIST_3               101032203
#define MON_PROCESS_PERSIST_4               101032204
#define MON_PROCESS_PERSIST_5               101032205
#define MON_PROCESS_PERSIST_6               101032206
#define MON_PROCESSCONT_KILLALLDOWN_1       101032301
#define MON_PROCESS_SETSTATE_1              101032401
#define MON_PROCESS_SETSTATE_2              101032402
#define MON_PROCESS_COMPLETESTARTUP_1       101032501
#define MON_PROCESS_COMPLETESTARTUP_2       101032502
#define MON_PROCESS_COMPLETESTARTUP_3       101032503
#define MON_PROCESSCONT_CHILDEXIT_1         101032601
#define MON_PROCESS_PROCEXITNOTIFIERNODES_1 101032701
#define MON_PROCESS_PROCEXIT_1              101032801
#define MON_PROCESS_PROCEXIT_2              101032802
#define MON_PROCESS_PROCEXITUNREGALL_1      101032901

/* Module: pnode.cxx = 04 */

#define MON_NODE_GETCPUSTAT_1               101040101
#define MON_NODE_GETSCHDATA_1               101040201
#define MON_NODE_GETSCHDATA_2               101040202
#define MON_NODE_GETSCHDATA_3               101040203
#define MON_NODE_GETSCHDATA_4               101040204
#define MON_NODE_GETSCHDATA_5               101040205
#define MON_NODE_GETSCHDATA_6               101040206
#define MON_NODE_GETSCHDATA_7               101040207
//#define MON_NODE_SETAFFINITY                101040301
#define MON_NODE_STARTWATCHDOG_1            101040501
#define MON_NODECONT_ASSIGN_LNODE           101040601
#define MON_NODECONT_LOAD_CONFIG_1          101040701
#define MON_NODECONT_LOAD_CONFIG_2          101040702
#define MON_NODECONT_LOAD_CONFIG_3          101040703
#define MON_NODECONT_LOAD_CONFIG_4          101040704
#define MON_NODE_SHUTDOWN_1                 101040801
#define MON_NODE_STARTPSTARTD_1             101040901
#define MON_NODE_STARTSMSERVICE_1           101041001
#define MON_NODE_ADDNODE_1                  101041101
#define MON_NODE_ADDNODE_2                  101041102
#define MON_NODE_ADDNODE_3                  101041103
#define MON_NODE_ADDLNODES_1                101041201
#define MON_NODE_ADDLNODES_2                101041202
#define MON_NODE_ADDLNODES_3                101041203
#define MON_NODE_ADDLNODES_4                101041204
#define MON_NODE_DELETENODE_1               101041301
#define MON_NODE_STARTNAMESERVER_1          101041401
#define MON_NODE_GETPROCESSNS_1             101041501
#define MON_NODE_GETPROCESSNS_2             101041502
#define MON_NODE_GETPROCESSNS_3             101041503
#define MON_NODE_GETPROCESSNS_4             101041504
#define MON_NODE_GETPROCESSNS_5             101041505
#define MON_NODE_GETPROCESSNS_6             101041506
#define MON_NODE_GETSTRINGID_1              101041601
#define MON_NODE_GETSTRINGID_2              101041602
#define MON_NODE_CLONEPROCESSNS_1           101041701
#define MON_NODE_CLONEPROCESSNS_2           101041702
#define MON_NODE_CLONEPROCESSNS_3           101041703
#define MON_NODE_CLONEPROCESSNS_4           101041704
#define MON_NODE_CLONEPROCESSNS_5           101041705
#define MON_NODE_CLONEPROCESSNS_6           101041706
#define MON_NODE_GETPROCESSLBYTYPENS_1      101041801
#define MON_NODE_GETPROCESSLBYTYPENS_2      101041802
#define MON_NODE_GETPROCESSLBYTYPENS_3      101041803

/* Module: config.cxx = 05 */

#define MON_CONFIGGROUP_SENDCHANGENOTICE_1  101050101
#define MON_CONFIGCONT_REPCOMPLETED_1       101050201
#define MON_CONFIGCONT_ADDUNIQUESTRING_1    101050301
#define MON_CONFIGCONT_INIT_1               101050401
#define MON_CONFIGCONT_INIT_2               101050402
#define MON_CONFIGCONT_INIT_3               101050403
#define MON_CONFIGCONT_INIT_4               101050404
#define MON_CONFIGCONT_INIT_5               101050405
#define MON_CONFIGCONT_INIT_6               101050406
#define MON_CONFIGCONT_ADDKEYNAME_1         101050501
#define MON_CONFIGCONT_ADDPROCNAME_1        101050601
#define MON_CONFIGCONT_ADDCLUSTERDATA_1     101050701
#define MON_CONFIGCONT_ADDPROCDATA_1        101050801
#define MON_CONFIGCONT_FINDUNIQUESTRING_1   101050901
#define MON_CONFIGCONT_GETUNIQUESTRIDMAX_1  101051001
#define MON_CONFIGCONT_STRINGIDTPSTRING_1   101051101

/* Module: tmsync.cxx = 06 */

#define MON_TMSYNC_BLOCK_SIG_1              101060101
#define MON_TMSYNC_UNBLOCK_SIG_1            101060201
#define MON_TMSYNC_INIT_1                   101060301
#define MON_TMSYNC_DEST_1                   101060401
#define MON_TMSYNC_UPDATE_STATE_1           101060501
#define MON_TMSYNC_COMMITTMDATA_1           101060601
#define MON_TMSYNC_END                      101060701
#define MON_TMSYNC_END_PENDING              101060801
#define MON_TMSYNC_Q_TMSYNC                 101060901
#define MON_TMSYNC_SEND_UNSOLICITED         101061001
#define MON_TMSYNC_UNPACKSYNCDATA           101061101
#define MON_TMSYNC_CANCEL_UNSOL_MESSAGE_1   101061201
#define MON_TMSYNC_UNSOL_COMPLETE_1         101061301
#define MON_TMSYNC_UNSOL_COMPLETE_2         101061302
#define MON_TMSYNC_UNSOL_COMPLETE_WAIT_1    101061401

/* Module: device.cxx = 07 */

#define MON_DEVICE_MOUNT_1                  101070101
#define MON_DEVICE_MOUNT_2                  101070102
#define MON_DEVICE_CONFIG_1                 101070201
#define MON_DEVICE_CONFIG_2                 101070202
#define MON_DEVICE_CONFIG_3                 101070203
#define MON_DEVICE_CONFIG_4                 101070204
#define MON_DEVICE_CONFIG_5                 101070205
#define MON_DEVICE_CONFIG_6                 101070206
#define MON_DEVICE_UNMOUNT_1                101070301
#define MON_DEVICE_UNMOUNT_2                101070302
#define MON_DEVICE_CONTAINER_1              101070401
#define MON_DEVICE_CONTAINER_2              101070402

/* Module: monsonar.cxx = 08 */

#define MON_SONAR_INIT_1                    101080101
#define MON_SONAR_INIT_2                    101080102

/* Module: clusterconf.cxx = 09 */

#define MON_CLUSTERCONF_INIT_1              101090101
#define MON_CLUSTERCONF_INIT_2              101090102
#define MON_CLUSTERCONF_INIT_3              101090103
#define MON_CLUSTERCONF_LOADNODE_1          101090201
#define MON_CLUSTERCONF_LOADNODE_2          101090202
#define MON_CLUSTERCONF_LOADPERSIST_1       101090301
#define MON_CLUSTERCONF_LOADPERSIST_2       101090302
#define MON_CLUSTERCONF_LOADPERSIST_3       101090303
#define MON_CLUSTERCONF_LOADPERSIST_4       101090304
#define MON_CLUSTERCONF_PROCESS_SNODE_1     101090401
#define MON_CLUSTERCONF_PROCESS_SNODE_2     101090402
#define MON_CLUSTERCONF_PROCESS_SNODE_3     101090403
#define MON_CLUSTERCONF_PROCESSPERSIST_1    101090501
#define MON_CLUSTERCONF_PROCESSPERSIST_2    101090502
#define MON_CLUSTERCONF_PROCESSPERSIST_3    101090503
#define MON_CLUSTERCONF_PROCESSPERSIST_4    101090504
#define MON_CLUSTERCONF_PROCESSPERSIST_5    101090505
#define MON_CLUSTERCONF_DELETENODE_1        101090601
#define MON_CLUSTERCONF_DELETEDBNODE_1      101090701
#define MON_CLUSTERCONF_DELETEDBNODE_2      101090702
#define MON_CLUSTERCONF_DELETEDBNODE_3      101090703
#define MON_CLUSTERCONF_DELETEDBNODE_4      101090704
#define MON_CLUSTERCONF_SAVENODE_1          101090801
#define MON_CLUSTERCONF_SAVELNODE_1         101090901
#define MON_CLUSTERCONF_SAVELNODE_2         101090902
#define MON_CLUSTERCONF_SAVEPNODE_1         101091001
#define MON_CLUSTERCONF_SAVEPNODE_2         101091002
#define MON_CLUSTERCONF_CLEAR_1             101091101
#define MON_CLUSTERCONF_UPDATEPNODECFG_1    101091201
#define MON_CLUSTERCONF_UPDATEDBPNODE_1     101091301
#define MON_CLUSTERCONF_UPDATEDBPNODE_2     101091302
#define MON_CLUSTERCONF_DELETEDBUSTRING_1   101091401
#define MON_CLUSTERCONF_DELETEDBUSTRING_2   101091402

/* Module: lock.cxx = 10 */

#define MON_LOCK_INIT_1                     101100101
#define MON_LOCK_INIT_2                     101100102
#define MON_LOCK_INIT_3                     101100103
#define MON_LOCK_INIT_4                     101100104
#define MON_LOCK_DEST_1                     101100201
#define MON_LOCK_DEST_2                     101100202
#define MON_LOCK_LOCK_1                     101100301
#define MON_LOCK_LOCK_2                     101100302
#define MON_LOCK_LOCK_3                     101100303
#define MON_LOCK_UNLOCK_1                   101100401
#define MON_LOCK_UNLOCK_2                   101100402
#define MON_LOCK_WAKEONE_1                  101100501
#define MON_LOCK_WAKEALL_1                  101100601
#define MON_LOCK_TMWAIT_1                   101100701
#define MON_LOCK_WAIT_1                     101100801
#define MON_LOCK_TRY_1                      101100901

/* Module: lnode.cxx = 11 */

#define MON_LNODE_SETAFFINITY_1             101110101
#define MON_LNODE_SETAFFINITY_2             101110102
#define MON_LNODE_MARKUP                    101110201

/* Module: mlio.cxx = 12 */

#define MON_MLIO_INIT_1                     101120101
#define MON_MLIO_INIT_2                     101120102
#define MON_MLIO_INIT_3                     101120103
#define MON_MLIO_INIT_4                     101120104
#define MON_MLIO_INIT_5                     101120105
#define MON_MLIO_INIT_6                     101120106
#define MON_MLIO_INIT_7                     101120107
#define MON_MLIO_INIT_8                     101120108
#define MON_MLIO_INIT_9                     101120109
#define MON_MLIO_INIT_10                    101120110
#define MON_MLIO_INIT_11                    101120111
#define MON_MLIO_INIT_12                    101120112
#define MON_MLIO_INIT_13                    101120113
#define MON_MLIO_DESTRUCT_1                 101120201
#define MON_MLIO_DESTRUCT_2                 101120202
#define MON_MLIO_MANAGE_NOTICE_1            101120301
#define MON_MLIO_SEND_NOTICE_1              101120401
#define MON_MLIO_SEND_NOTICE_2              101120402
#define MON_MLIO_SEND_NOTICE_3              101120403
#define MON_MLIO_SENDCTL_1                  101120501
#define MON_MLIO_PROCESS_TH_1               101120601
#define MON_MLIO_PROCESS_TH_2               101120602
#define MON_MLIO_SERIALREQ_TH_1             101120701
#define MON_MLIO_SERIALREQ_TH_2             101120702
#define MON_MLIO_SERIALREQ_TH_3             101120703
#define MON_MLIO_TMSYNC_TH_1                101120801
#define MON_MLIO_TMSYNC_TH_2                101120802
#define MON_MLIO_PENDING_TH_1               101120901
#define MON_MLIO_INIT_WORKER_1              101121001
#define MON_MLIO_INIT_WORKER_2              101121002
#define MON_MLIO_INIT_WORKER_3              101121003
#define MON_MLIO_INIT_WORKER_4              101121004
#define MON_MLIO_ACQUIRE_MSG_1              101121101
#define MON_MLIO_ACQUIRE_MSG_2              101121102
#define MON_MLIO_RELEASE_MSG_1              101121201
#define MON_MLIO_QSTATS_1                   101121301
#define MON_MLIO_SHUTDOWN_1                 101121401
#define MON_MLIO_SHUTDOWN_2                 101121402
#define MON_MLIO_NOTICE_DEST_1              101121501
#define MON_MLIO_NOTICE_DEST_2              101121502
#define MON_MLIO_DEADPID_OVERFLOW           101121503

/* Module: redirect.cxx = 13 */

#define MON_REDIR_REDIR_1                   101130101
#define MON_REDIR_REDIR_2                   101130102
#define MON_REDIR_UREDIR_1                  101130201
#define MON_REDIR_KILL_ERR                  101130301
#define MON_REDIR_DISPOSEIODATA_1           101130401
#define MON_REDIRECT_TH_1                   101130501
#define MON_REDIRECT_TH_2                   101130502
#define MON_REDIRECT_TH_3                   101130503
#define MON_REDIRECT_TH_4                   101130504
#define MON_REDIRECT_TH_5                   101130505
#define MON_REDIRECT_TH_6                   101130506
#define MON_REDIR_STDIN_FD_1                101130601
#define MON_REDIR_STDIN_FD_2                101130602
#define MON_REDIR_SHUTDOWNWORK              101130701
#define MON_REDIR_STDERR                    101130801
#define MON_REDIR_SETFDFLAGS_1              101130901
#define MON_REDIR_STDIN_TTY_1               101131001
#define MON_REDIR_USTDIN_TTY_1              101131101
#define MON_REDIR_STDIN_TTY_HNDLIN_1        101131201
#define MON_REDIR_STDIN_TTY_HNDLOUT_1       101131301
#define MON_REDIR_STDIN_FILE_1              101131401
#define MON_REDIR_USTDIN_FILE_1             101131501
#define MON_REDIR_STDIN_HNDLIN_1            101131601
#define MON_REDIR_STDIN_HNDLIN_2            101131602
#define MON_REDIR_STDIN_HNDLOUT_1           101131701
#define MON_REDIR_ANCSTDIN_HNDLIN_1         101131801
#define MON_REDIR_STDINREMOTE_1             101131901
#define MON_REDIR_STDINREMOTE_2             101131902
#define MON_REDIR_USTDINREMOTE_1            101132001
#define MON_REDIR_USTDINREM_HNDLIN_1        101132101
#define MON_REDIR_STDOUT_1                  101132201
#define MON_REDIR_USTDOUT_1                 101132301
#define MON_REDIR_STDOUT_HNDLOUT_1          101132401
#define MON_REDIR_ADDTOEPOLL_1              101132501
#define MON_REDIR_DELFROMEPOLL_1            101132601
#define MON_REDIR_STDINON_1                 101132701
#define MON_REDIR_SHUTPIPE_FD_1             101132801
#define MON_REDIR_REDIRECT_1                101132901
#define MON_REDIR_START_1                   101133001
#define MON_REDIR_LOGWRITE                  101133101


/* Module: lnodeconf.cxx = 14 */

#define MON_LNODECONF_CONSTR_1              101140101
#define MON_LNODECONF_ADD_LNODE_1           101140201
#define MON_LNODECONF_ADD_LNODE_2           101140202

/* Module: pnodeconf.cxx = 15 */

#define MON_PNODECONF_SET_SPARE_1           101150101
#define MON_PNODECONF_CONSTR_1              101150201
#define MON_PNODECONF_ADD_PNODE_1           101150301
#define MON_PNODECONF_ADD_PNODE_2           101150302

/* Module: reqworker.cxx = 16 */

#define MON_REQWORKER_1                     101160101
#define MON_REQWORKER_2                     101160102
#define MON_REQWORKER_SHUTDOWN_1            101160201

/* Module: intprocess.cxx = 17 */

#define MON_INTPROCESS_1                    101170101
#define MON_INTPROCESS_2                    101170102
#define MON_INTPROCESS_3                    101170103
#define MON_INTPROCESS_4                    101170104
#define MON_INTPROCESS_5                    101170105

/* Modules: reqqueue.cxx req<xxxxx>.cxx = 18 */

#define MON_REQQUEUE_PREP_REQ_1             101180101
#define MON_REQQUEUE_PREP_EXT_REQ_1         101180201
#define MON_REQQUEUE_QUIESCE_1              101180301
#define MON_REQQUEUE_SNAPSHOT_1             101180401
#define MON_REQQUEUE_SNAPSHOT_2             101180402
#define MON_REQQUEUE_SNAPSHOT_3             101180403
#define MON_REQQUEUE_SNAPSHOT_4             101180404
#define MON_REQQUEUE_SNAPSHOT_5             101180405
#define MON_REQQUEUE_SNAPSHOT_6             101180406
#define MON_REQQUEUE_SNAPSHOT_7             101180407
#define MON_REQQUEUE_SNAPSHOT_8             101180408
#define MON_REQQUEUE_SNAPSHOT_9             101180409
#define MON_REQQUEUE_SNAPSHOT_10            101180410
#define MON_REQQUEUE_SNAPSHOT_11            101180411
#define MON_REQQUEUE_SNAPSHOT_12            101180412
#define MON_REQQUEUE_SNAPSHOT_13            101180413
#define MON_REQQUEUE_SNAPSHOT_14            101180414
#define MON_REQQUEUE_SNAPSHOT_15            101180415
#define MON_REQQUEUE_REVIVE_1               101180501
#define MON_REQQUEUE_REVIVE_2               101180502
#define MON_REQQUEUE_REVIVE_3               101180503
#define MON_REQQUEUE_REVIVE_4               101180504
#define MON_REQQUEUE_REVIVE_5               101180505
#define MON_REQQUEUE_REVIVE_6               101180506
#define MON_REQQUEUE_REQUEST_1              101180601
#define MON_REQ_NODE_ADD_1                  101180701
#define MON_REQ_NODE_ADD_2                  101180702
#define MON_REQ_NODE_ADD_3                  101180703
#define MON_REQ_NODE_ADD_4                  101180704
#define MON_REQ_NODE_DELETE_1               101180801
#define MON_REQ_NODE_DELETE_2               101180802
#define MON_REQ_NODE_DELETE_3               101180803
#define MON_REQ_NODE_DELETE_4               101180804
#define MON_REQ_NODE_NAME_1                 101180901
#define MON_REQ_NODE_NAME_2                 101180902
#define MON_INTREQ_NODE_NAME_1              101181001
#define MON_INTREQ_NODE_NAME_2              101181002
#define MON_INTREQ_NODE_NAME_3              101181003
#define MON_REQ_NAMESERVER_ADD_1            101181101
#define MON_REQ_NAMESERVER_ADD_2            101181102
#define MON_REQ_NAMESERVER_ADD_3            101181103
#define MON_REQ_NAMESERVER_ADD_4            101181104
#define MON_REQ_NAMESERVER_DELETE_1         101181201
#define MON_REQ_NAMESERVER_DELETE_2         101181202
#define MON_REQ_NAMESERVER_DELETE_3         101181203
#define MON_REQ_NAMESERVER_DELETE_4         101181204
#define MON_REQ_NAMESERVER_INFO_1           101181301
#define MON_REQ_EVALREQ_PERFORMANCE_1       101181501
#define MON_INTREQ_CLONEPROC_1              101181601
#define MON_INTREQ_EXIT_1                   101181701
#define MON_INTREQ_NEWPROC_1                101181801
#define MON_INTREQ_NEWPROC_2                101181802
#define MON_INTREQ_NEWPROC_3                101181803
#define MON_INTREQ_NOTIFY_1                 101181901
#define MON_INTREQ_NOTIFY_2                 101181902
#define MON_INTREQ_NOTIFY_3                 101181903
#define MON_INTREQ_NOTIFY_4                 101181904
#define MON_INTREQ_OPEN_1                   101182001
#define MON_REQ_IODATA_1                    101182201
#define MON_REQ_STDIN_1                     101182301
#define MON_REQ_STDIN_2                     101182302
#define MON_REQ_KILL_1                      101182401
#define MON_REQ_NOTIFY_1                    101182501
#define MON_REQ_PROCINFO_1                  101182601
#define MON_REQ_PROCINFOCONT_1              101182701
#define MON_INTREQ_CHILDDEATH_1             101182801

/* Module: clio.cxx = 19 */
#define MON_CLIO_ACQUIRE_MSG_1              101190101
#define MON_CLIO_ACQUIRE_LOCK_1             101190201
#define MON_CLIO_ACQUIRE_LOCK_2             101190202
#define MON_CLIO_RELEASE_LOCK_1             101190301
#define MON_CLIO_RELEASE_LOCK_2             101190302
#define MON_CLIO_RELEASE_MSG_1              101190401
#define MON_CLIO_RELEASE_MSG_2              101190402
#define MON_CLIO_LEAK_1                     101190501


/* Module: cmsg.cxx = 20 */
#define MON_CMSH_GET_CLUSTER_STATE_1        101200101
//#define MON_CMSH_GET_CLUSTER_STATE_2        101200102

/* Module: system.cxx = 21 */
#define MON_SYSTEM_LAUNCH_COMMAND_1         101210101
#define MON_SYSTEM_LAUNCH_COMMAND_2         101210102
#define MON_SYSTEM_LAUNCH_COMMAND_3         101210103
#define MON_SYSTEM_WAIT_COMMAND_1           101210201
#define MON_SYSTEM_WAIT_COMMAND_2           101210202
#define MON_SYSTEM_WAIT_COMMAND_3           101210203
#define MON_UTILITY_EXECUTECOMMAND_1        101210301

/* Module: watchdog.cxx = 22 */
#define MON_WATCHDOG_MAIN_1                 101220101
#define MON_WATCHDOG_MAIN_2                 101220102
#define MON_WATCHDOG_MAIN_3                 101220103
#define MON_WATCHDOG_MAIN_4                 101220104
#define MON_WATCHDOG_MAIN_5                 101220105
#define MON_WATCHDOG_MAIN_6                 101220106
#define MON_WATCHDOG_MAIN_7                 101220107
#define MON_WATCHDOG_PROCSTARTUP_1          101220201
#define MON_WATCHDOG_INITLOCALIO_1          101220301
#define MON_WATCHDOG_INITLOCALIO_2          101220302
#define MON_WATCHDOG_INITLOCALIO_3          101220303

/* Module: healthcheck.cxx = 23 */
#define MON_HEALTHCHECK_BAD_STATE           101230101
#define MON_HEALTHCHECK_HC_1                101230102
#define MON_HEALTHCHECK_START_1             101230103
#define MON_HEALTHCHECK_QUIESCE_1           101230104
#define MON_HEALTHCHECK_TEVENT_1            101230201
#define MON_HEALTHCHECK_TEVENT_2            101230202
#define MON_HEALTHCHECK_TEVENT_3            101230203
#define MON_HEALTHCHECK_TEVENT_4            101230204
#define MON_HEALTHCHECK_SCH_1               101230301
#define MON_HEALTHCHECK_WAKEUP_1            101230401
#define MON_HEALTHCHECK_WAKEUP_2            101230402
#define MON_HEALTHCHECK_Q_BLOCK             101230501

#define MON_HEALTHCHECK_STOP_NS_1           101230701

/* Module: sdtimer.cxx = 24 */
#define MON_SDTIMER_SOFTDOG_TH_1            101240101
#define MON_SDTIMER_MONITORINDEBUG_1        101240201
#define MON_SDTIMER_NODEFAILSAFE_1          101240301
#define MON_SDTIMER_NODEFAILSAFE_2          101240302
#define MON_SDTIMER_NODEFAILSAFE_3          101240303
#define MON_SDTIMER_STARTWORKER_1           101240401
#define MON_SDTIMER_STOPPROCESSES_1         101240501
#define MON_SDTIMER_STOPMONITORPROC_1       101240601
#define MON_SDTIMER_SHUTDOWNWORK_1          101240701
#define MON_SDTIMER_SHUTDOWNWORK_2          101240702
#define MON_SDTIMER_UNMOUNTDEVS_1           101240801
#define MON_SDTIMER_SUSPENDMONITORPROC_1    101240901
#define MON_SDTIMER_TIMEREXPIRED_1          101241001

/* Module: wdtimer.cxx = 25 */
#define MON_WDTIMER_WATCHDOG_TH_1           101250101
#define MON_WDTIMER_STARTWORKER_1           101250201
#define MON_WDTIMER_SHUTDOWNWORK_1          101250301
#define MON_WDTIMER_SHUTDOWNWORK_2          101250302
#define MON_WDTIMER_RESET_WATCHTIMER        101250401
#define MON_WDTIMER_RESTORE_WATCHTIMER_1    101250501
#define MON_WDTIMER_RESTORE_WATCHTIMER_2    101250502
#define MON_WDTIMER_SETMIN_WATCHTIMER_1     101250601
#define MON_WDTIMER_SETMIN_WATCHTIMER_2     101250602
#define MON_WDTIMER_START_WATCHTIMER_1      101250701
#define MON_WDTIMER_START_WATCHTIMER_2      101250702
#define MON_WDTIMER_START_WATCHTIMER_3      101250703
#define MON_WDTIMER_START_WATCHTIMER_4      101250704
#define MON_SDTIMER_DUMPMONITORPROC_1       101250801

/* Module: lunmgr.cxx = 26 */
#define MON_LUNMGR_LAUNCHCOMMAND_1          101260101
#define MON_LUNMGR_WAITCOMMAND_1            101260201

/* Module: lunmgr.cxx = 27 */
#define MON_PKILLALL_EXECUTECOMMAND_1       101270101

/* Module: procmon.cxx = 28 */
#define MON_PROCMON_PROCESSMONITOR_TH_1     101280101
#define MON_PROCMON_STARTWORKER_1           101280101
#define MON_PROCMON_SHUTDOWNWORK_1          101280201
#define MON_PROCMON_CHECKLUNMGR_1           101280301
#define MON_PROCMON_CHECKMONITOR_1          101280401

/* Module monlogging.cxx = 29 */
#define MON_LOG_ERROR_1                     101290101
#define MON_LOG_ERROR_2                     101290102
#define MON_LOG_ERROR_3                     101290103

/* Module pstartd.cxx = 30 */
#define LIO_NOTICE_CALLBACK_1               101300101
#define LIO_NOTICE_CALLBACK_2               101300102
#define LIO_EVENT_CALLBACK_1                101300201
#define LIO_EVENT_CALLBACK_2                101300203
#define MONUTIL_REQUEST_GET_1               101300301
#define MONUTIL_REQUEST_GET_2               101300303
#define MONUTIL_REQUEST_EXIT_1              101300401
#define MONUTIL_REQUEST_EXIT_2              101300402
#define MONUTIL_REQUEST_EXIT_3              101300403
#define MONUTIL_REQUEST_EXIT_4              101300404
#define MONUTIL_REQUEST_NEWPROC_1           101300501
#define MONUTIL_REQUEST_NEWPROC_2           101300502
#define MONUTIL_REQUEST_NEWPROC_3           101300503
#define MONUTIL_REQUEST_NEWPROC_4           101300504
#define MONUTIL_REQUEST_PROCINFO_1          101300601
#define MONUTIL_REQUEST_PROCINFO_2          101300602
#define MONUTIL_REQUEST_PROCINFO_3          101300603
#define MONUTIL_REQUEST_PROCINFO_4          101300604
#define MONUTIL_REQUEST_STARTUP_1           101300701
#define MONUTIL_PERFORM_REQUEST_1           101300801
#define PSTARTD_PSTARTD_1                   101300901
#define PSTARTD_START_PROCESS_1             101300901
#define PSTARTD_STARTPROCS_1                101301001
#define PSTARTD_STARTPROCS_2                101301002
#define PSTARTD_STARTPROCS_3                101301003
#define PSTARTD_MAIN_1                      101301101


/* Module robsem.cxx = 31 */
#define MON_ROBSEM_1                        101310101
#define MON_ROBSEM_2                        101310102
#define MON_ROBSEM_3                        101310103
#define MON_ROBSEM_4                        101310104

/* Module commaccept.cxx = 32 */
#define MON_COMMACCEPT_1                    101320101
#define MON_COMMACCEPT_2                    101320102
#define MON_COMMACCEPT_3                    101320103
#define MON_COMMACCEPT_4                    101320104
#define MON_COMMACCEPT_5                    101320105
#define MON_COMMACCEPT_6                    101320106
#define MON_COMMACCEPT_7                    101320107
#define MON_COMMACCEPT_8                    101320108
#define MON_COMMACCEPT_9                    101320109
#define MON_COMMACCEPT_10                   101320110
#define MON_COMMACCEPT_11                   101320111
#define MON_COMMACCEPT_12                   101320112
#define MON_COMMACCEPT_13                   101320113
#define MON_COMMACCEPT_14                   101320114
#define MON_COMMACCEPT_15                   101320115
#define MON_COMMACCEPT_16                   101320116
#define MON_COMMACCEPT_17                   101320117
#define MON_COMMACCEPT_18                   101320118
#define MON_COMMACCEPT_19                   101320119

/* Module: reqnodedown.cxx = 33 */
#define MON_EXT_NODEDOWN_REQ                101330101

/* Module: gentrap.cxx = 34 */
#define MON_GEN_SNMP_TRAP                   101340101

/* Module: persistconfig.cxx = 36 */
#define MON_PERSISTCONFIG_ADDCONFIG_1       101360101

/* Module: zclient.cxx = 37 */
#define MON_ZCLIENT_ZCLIENT_1               101370101
#define MON_ZCLIENT_ZCLIENT_2               101370102
#define MON_ZCLIENT_ZCLIENT_3               101370103
#define MON_ZCLIENT_SYNC_STRING_COMP_1      101370201
#define MON_ZCLIENT_CHECKCLUSTER_1          101370301
#define MON_ZCLIENT_CHECKCLUSTER_2          101370302
#define MON_ZCLIENT_CHECKCLUSTERZNODES_1    101370401
#define MON_ZCLIENT_CHECKCLUSTERZNODES_2    101370402
#define MON_ZCLIENT_CHECKCLUSTERZNODES_3    101370403
#define MON_ZCLIENT_GETCLUSTERZNODES_1      101370501
#define MON_ZCLIENT_GETCLUSTERZNODES_2      101370502
#define MON_ZCLIENT_REGISTERZNODE_1         101370601
#define MON_ZCLIENT_SHUTDOWNWORK_1          101370701
#define MON_ZCLIENT_ZCLIENTTHREAD_1         101370801
#define MON_ZCLIENT_STARTWORK_1             101370901
#define MON_ZCLIENT_MONITORZCLUSTER_1       101371001
#define MON_ZCLIENT_GETZNODEDATA_1          101371101
#define MON_ZCLIENT_GETZNODEDATA_2          101371102
#define MON_ZCLIENT_GETZNODEDATA_3          101371103
#define MON_ZCLIENT_WATCHCLUSTER_1          101371201
#define MON_ZCLIENT_WATCHCLUSTER_2          101371202
#define MON_ZCLIENT_SETZNODEWATCH_1         101371301
#define MON_ZCLIENT_SETZNODEWATCH_2         101371302
#define MON_ZCLIENT_WATCHNODE_1             101371401
#define MON_ZCLIENT_ZSESSIONWATCHER_1       101371501
#define MON_ZCLIENT_ZSESSIONWATCHER_2       101371502
#define MON_ZCLIENT_CHECKZNODE_1            101371601
#define MON_ZCLIENT_WATCHNODEDELETE_1       101371701
#define MON_ZCLIENT_WATCHNODEDELETE_2       101371702
#define MON_ZCLIENT_WATCHNODEDELETE_3       101371703
#define MON_ZCLIENT_ISZNODEEXPIRED_1        101371801
#define MON_ZCLIENT_ISZNODEEXPIRED_2        101371802
#define MON_ZCLIENT_CHECKMYZNODE_1          101371901
#define MON_ZCLIENT_CHECKMYZNODE_2          101371902
#define MON_ZCLIENT_AMICONFIGUREDMASTER_1   101372101
#define MON_ZCLIENT_AMICONFIGUREDMASTER_2   101372102
#define MON_ZCLIENT_WAITFORANDRETURNMASTER  101372103
#define MON_ZCLIENT_CREATEMASTERZNODE       101372104
#define MON_ZCLIENT_WATCHMASTERNODEDELETE_1 101372105
#define MON_ZCLIENT_WATCHMASTERNODEDELETE_2 101372106
#define MON_ZCLIENT_WATCHMASTERNODEDELETE_3 101372107
#define MON_ZCLIENT_CREATEORSETMASTERWATCH  101372108
#define MON_ZCLIENT_CREATEORSETMASTERINFO   101372109

/* Module: zconfig.cxx = 38 */
#define ZCONFIG_ZCONFIG_1                   101380101
#define ZCONFIG_ZCONFIG_2                   101380102
#define ZCONFIG_ZCONFIG_3                   101380103
#define ZCONFIG_ZCONFIG_4                   101380104
#define ZCONFIG_CHECKCONFIG_1               101380201
#define ZCONFIG_CHECKCONFIG_2               101380202
#define ZCONFIG_GETCONFIGZNODESCHILDREN_1   101380301
#define ZCONFIG_GETZNODEDATA_1              101380401
#define ZCONFIG_GETZNODEDATA_2              101380402
#define ZCONFIG_ISZNODEEXISTENT_1           101380501
#define ZCONFIG_ISZNODEEXISTENT_2           101380502
#define ZCONFIG_MAKECONFIGZNODE_1           101380601
#define ZCONFIG_REGISTERZNODE_1             101380701
#define ZCONFIG_SETZNODEWATCH_1             101380801
#define ZCONFIG_SETZNODEWATCH_2             101380802
#define ZCONFIG_WATCHCONFIG_1               101380901
#define ZCONFIG_WATCHCONFIG_2               101380902
#define ZCONFIG_WATCHCONFIG_2               101380902
#define ZCONFIG_WATCHCONFIGZNODE_1          101381001
#define ZCONFIG_DELETECONFIGZNODE_1         101381001
#define ZCONFIG_DELETECONFIGZNODE_2         101381002
#define ZCONFIG_DELETECONFIGZNODE_3         101381003

/* Module nameserver.cxx = 39 */
#define NAMESERVER_CLIENTSOCKCREATE_1       101390101
#define NAMESERVER_CLIENTSOCKCREATE_2       101390102
#define NAMESERVER_CLIENTSOCKCREATE_3       101390103
#define NAMESERVER_CLIENTSOCKCREATE_4       101390104
#define NAMESERVER_CLIENTSOCKCREATE_5       101390105
#define NAMESERVER_CLIENTSOCKCREATE_6       101390106
#define NAMESERVER_SENDTONS_1               101390201
#define NAMESERVER_SENDTONS_2               101390202
#define NAMESERVER_SOCKRECEIVE_1            101390301
#define NAMESERVER_SOCKSEND_1               101390401
#define NAMESERVER_GETM2NPORT_1             101390501
#define NAMESERVER_CHOOSENEXTNS_1           101390601

/* Module nscommaccept.cxx = 40 */
#define NS_COMMACCEPT_PROCESSMONREQS_1      101400101
#define NS_COMMACCEPT_PROCESSMONREQS_2      101400102
#define NS_COMMACCEPT_PROCESSMONREQS_3      101400103
#define NS_COMMACCEPT_PROCESSMONREQS_4      101400104
#define NS_COMMACCEPT_PROCESSMONREQS_5      101400105
#define NS_COMMACCEPT_PROCESSNEWSOCK_1      101400201
#define NS_COMMACCEPT_COMMACCEPTORSOCK_1    101400301
#define NS_COMMACCEPT_MON2NSACCEPTMON_1     101400401
#define NS_COMMACCEPT_MON2NSPROCESS_1       101400501
#define NS_COMMACCEPT_START_1               101400601

/* Module: reqnodedown.cxx = 41 */
#define MON_EXT_NAMESERVERDOWN_REQ          101410101

/* Module: nsreqnewproc.cxx = 42 */
#define NS_EXTNEWPROCNSREQ_1                101420101
#define NS_EXTNEWPROCNSREQ_2                101420102

/* Module: tcdb.cxx = 90 */
#define TCDB_TCDB_1                         101900101
#define TCDB_TCDB_2                         101900102
#define TCDB_TCDB_3                         101900103

/* Module: tcdbzstore.cxx = 91 */
#define TCDB_ZSTORE_SESSIONWATCHER_1        101910101
#define TCDB_ZSTORE_SESSIONWATCHER_2        101910102
#define TCDB_ZSTORE_ZSTORE_1                101910201
#define TCDB_ZSTORE_ZSTORE_2                101910202
#define TCDB_ZSTORE_ZSTORE_3                101910203
#define TCDB_ZSTORE_ZSTORE_4                101910204
#define TCDB_ZSTORE_MAKECLUSTERZNODES_1     101910301
#define TCDB_ZSTORE_MAKECLUSTERZNODES_2     101910302
#define TCDB_ZSTORE_MAKECLUSTERZNODES_3     101910303

/* Module: tcdbsqlite.cxx = 92 */
#define SQLITE_DB_ACCESS_ERROR              101920101

/* Module ptpclient.cxx = 93 */
#define PTPCLIENT_PTPCLIENT_1               101930101
#define PTPCLIENT_STDINREQ_1                101930201
#define PTPCLIENT_STDINREQ_2                101930202
#define PTPCLIENT_STDIODATA_1               101930301
#define PTPCLIENT_STDIODATA_2               101930302
#define PTPCLIENT_SENDTOMON_1               101930401
#define PTPCLIENT_SENDTOMON_2               101930402

/* Module ptpcommaccept.cxx = 94 */
#define PTP_COMMACCEPT_1                    101940101
#define PTP_COMMACCEPT_2                    101940102
#define PTP_COMMACCEPT_3                    101940103
#define PTP_COMMACCEPT_4                    101940104
#define PTP_COMMACCEPT_5                    101940105
#define PTP_COMMACCEPT_6                    101940106
#define PTP_COMMACCEPT_7                    101940107
#define PTP_COMMACCEPT_8                    101940108

/* Module notice.cxx = 95 */
#define NOTICE_NOTIFYREMOTE_1               101950101

/**********************************************/

/*********** Seabed ***********/
#define SB_EVENT_ID        102004001


/**********************************************/
/**********************************************/


/*********** DTM ***********/
#define DTM_INVALID_SYNC_DATA 103005001
#define DTM_TAKEOVER_TX_FAILED 103005002
#define DTM_INVALID_MESSAGE_TYPE 103005003
#define DTM_CANNOT_OPEN_DTM 103005004
#define DTM_MSG_TO_DTM_FAILED 103005005
#define DTM_MSG_TO_DTM_FAILED2 103005006
#define DTM_RM_CLOSED 103005007
#define DTM_COMMIT_HEURISTIC 103005008
#define DTM_COMMIT_INCONSISTENT 103005009
#define DTM_ROLLBACK_HEURISTIC 103005010
#define DTM_RM_OPEN_FAILED 103005011
#define DTM_SYNC_SEND_FAILED 103005012
#define DTM_XATM_RECOVER_FAILED 103005013
#define DTM_ABORT_FAILED 103005014
#define DTM_CANNOT_OPEN_AMP 103005015
#define DTM_MSG_TO_AMP_FAILED 103005016
#define DTM_XARM_ATTACH_FAILED 103005017
#define DTM_XARM_STARTUP_FAILED 103005018
#define DTM_RECOVERY_SCAN_ERROR 103005019
#define DTM_HBASE_INIT_FAILED 103005020
#define DTM_TMPOOL_THRESHOLD_EXCEEDED 103005021
#define DTM_RMINFOHBASE_NULLTXNID 103005022
#define DTM_RECOV_FAIL_BAD_TXN_TYPE 103005023
#define DTM_TXN_INIT_RMS_NO_BRANCH_OBJ 103005024
#define DTM_NO_MORE_SEQNUMS 103005025
#define DTM_SEQNUM_COLLISION 103005026
#define DTM_SEQNUM_WRAPAROUND 103005027
#define DTM_BEGINTRANSACTION_COLLISION 103005028
#define DTM_NODE_RECOVERY_INITIATED 103005029
#define DTM_RECOVERY_COMPLETED 103005030
#define DTM_RM_REPLY_FAILED 103005031
#define DTM_RM_OPEN_FAILED1 103005032
#define DTM_RECOVERY_FAILED1 103005033
#define DTM_RECOVERY_FAILED2 103005034
#define DTM_RECOVERY_FAILED3 103005035
#define DTM_INVALID_MESSAGE_TYPE2 103005036
#define DTM_TM_INTEGRATING_TSE 103005037
#define DTM_TM_TSE_FAILOVER 103005038
#define DTM_MAX_THREADS 103005039
#define DTM_OUT_OF_MEMORY 103005040
#define DTM_OVERRIDE_AUDIT_INCONSISTENCY_WARN 103005041
#define DTM_RM_NOT_FOUND 103005042
#define DTM_RM_NOT_FOUND2 103005043
#define DTM_RM_FAILURE_DOWNED_TM 103005044
#define DTM_TM_STATE_CHANGE 103005045
#define DTM_RECOVERY_TXNS_TO_RECOVER 103005046
#define DTM_RECOV_BAD_STATE 103005047
#define DTM_RECOV_BAD_TYPE 103005048
#define DTM_RECOV_TRANSACTION_COLLISION 103005049
#define DTM_TT_READONLY_BROKEN 103005050
#define DTM_TIMER_TH_WAITING_FOR_EXIT 103005051
#define DTM_TM_INFO_SENDALL_FAIL 103005052
#define DTM_TM_ENABLETRANS_FAIL 103005053
#define DTM_TM_DISABLETRANS_FAIL 103005054
#define DTM_TM_LINK_PAUSED 103005055
#define DTM_TM_UNEXPECTED_CP_RECEIVED 103005056
#define DTM_TM_UNEXPECTED_PS_RECEIVED 103005057
#define DTM_TM_LEADTM_BAD 103005058
#define DTM_TM_SHUTDOWNP1WAIT_ERR 103005059
#define DTM_TM_SHUTDOWNP1WAIT_TXNS 103005060
#define DTM_TM_DISABLETRANS_TOOLATE 103005061
#define DTM_TM_TAKEOVER_FAILED 103005062
#define DTM_TM_TAKEOVER_NOMEM 103005063
#define DTM_TM_LINK_PAUSED2 103005064
#define DTM_AUDIT_FAILED_ADPOPEN 103005065
#define DTM_AX_REG_NID_PID_BAD 103005066
#define DTM_AX_REG_NOTFOUND 103005067
#define DTM_AX_REG_PTYPE_BAD 103005068
#define DTM_RECOV_NOT_LEAD_TM 103005069
#define DTM_RECOV_LOST_TM_DURING_SYSRECOV 103005070
#define DTM_RECOV_DUP_TXN 103005071
#define DTM_RM_REPLIES_OUTSTANDING 103005072
#define DTM_RM_OPEN_FAILED2 103005073
#define DTM_RECOVERY_FAILED4 103005074
#define DTM_TM_INTEGRATING_TSE2 103005075
#define DTM_TM_DRAINTRANS_TOOLATE 103005076
#define DTM_TM_QUIESCED 103005077
#define DTM_TM_ADD_APP_PARTIC_NOMEM 103005078
#define DTM_SHUTDOWN_DIRTY 103005079
#define DTM_SHUTDOWN_CLEAN 103005080
#define DTM_RECOV_INDOUBT_TXNS 103005081
#define DTM_RECOV_SYSRECOV_COMPLETE 103005082
#define DTM_RECOV_RESOLVING_TXN 103005083
#define DTM_TMTX_FORGET_HEURISTIC 103005084
#define DTM_TMTX_REG_TT_FLAGS_CHANGED 103005085
#define DTM_TM_NO_TRANS 103005086
#define DTM_NOUNDO 103005087
#define DTM_TMTX_DOOMTXN_ADDED_PARTIC 103005088
#define DTM_RMIDLOOKUP_NID_PID_BAD 103005089
#define DTM_DOOMTXN_NID_PID_BAD 103005090
#define DTM_AX_REG_XARM_NOTSUPPORTED 103005091
#define DTM_TMTX_REG_ERROR_TOOLATE 103005092
#define DTM_TMTX_ALREADY_PARTIC 103005093
#define DTM_PROCDEATH_TSE 103005094
#define DTM_PROCDEATH_DTM 103005095
#define DTM_NODEDOWN 103005096
#define DTM_NODEUP 103005097
#define DTM_NODEQUIESCE 103005098
#define DTM_XATM_RECOVER_SEND_FAILED 103005099
#define DTM_XATM_RECOVER_SEND2_FAILED 103005100
#define DTM_XATM_XA_RECOVER_TIMEOUT 103005101
#define DTM_XATM_COMPLETEALL_FAILED 103005102
#define DTM_XATM_SENDTORM_FAILED 103005103
#define DTM_XATM_RETRY_RM_SEND 103005104
#define DTM_XATM_CNT_EXCD_LMT 103005105
#define DTM_XATM_PROG_ERROR 103005106
#define DTM_XATM_XA_OPEN_FAILED 103005107
#define DTM_XATM_COMPLETEONE_FAILED 103005108
#define DTM_XATM_MAX_RM_MSG 103005109
#define DTM_LOGIC_ERROR_RM_OBJ 103005110
#define DTM_XATM_FAILED_RETRIES_EXC 103005111
#define DTM_XATM_RECOVER_FAILED_UNEXP_TAG 103005112
#define DTM_XATM_RECOVER_FAILED_PROG_ERROR 103005113
#define DTM_XATM_MSG_INTEGRITY_FAILED1 103005114
#define DTM_XATM_MSG_INTEGRITY_FAILED2 103005115
#define DTM_XATM_MSG_INTEGRITY_FAILED3 103005116
#define DTM_XATM_RECOVER_FAILED_UNEXP_REPLY 103005117
#define DTM_XATM_SENDTORM_FAILED2 103005118
#define DTM_XATM_OPEN_TSE 103005119
#define DTM_XATM_OPEN_GENERIC 103005120
#define DTM_XATM_RECOVER_WAIT_FAILED 103005121
#define DTM_XATM_RECOVER_SENDALL_FAILED 103005122
#define DTM_XATM_RECOVER_SENDALL2_FAILED 103005123
#define DTM_XATM_RECOVER_SENDALL_WAIT_FAILED 103005124
#define DTM_TSE_FAILURE_DETECTED 103005125
#define DTM_TSE_FAILOVER_DETECTED 103005126
#define DTM_PROCDEATH_ASE 103005127
#define DTM_REGCHANGE_NOTICE 103005128
#define DTM_SHUTDOWN_NOTICE 103005129
#define DTM_DISABLE_TRANSACTIONS_NORMAL 103005130
#define DTM_DISABLE_TRANSACTIONS_IMMEDIATE 103005131
#define DTM_TM_RESTART_RETRY_PROGERROR 103005132
#define DTM_TMRECOV_BAD_RMID 103005133
#define DTM_ERROR_SHUTDOWN_ABRUPT 103005134
#define DTM_ERROR_SHUTDOWN_DEBUG 103005135
#define DTM_TM_RECOVERED 103005136
#define DTM_XATM_RECOVER_SENDALL_WAIT_XAER_RMFAIL 103005137
#define DTM_RECOVERY_SCAN_ALL1 103005138
#define DTM_RECOVERY_SCAN_ALL2 103005139
#define DTM_TM_PROCESS_STARTUP 103005140
#define DTM_SYSTEM_RECOVERY_INITIATED 103005141
#define DTM_RM_RECOVERY_INITIATED 103005142
#define DTM_NODEPREPARE 103005143
#define DTM_TM_LISTBUILT_SYNC 103005144
#define DTM_TM_PROCESS_RESTART_SYNC 103005145
#define DTM_TM_GET_NODEINFO_FAILED 103005146
#define DTM_LEAD_TM_TM_SYNC_UNEXPECTED 103005147
#define DTM_BROADCAST_ROLLBACKS_INFO 103005148
#define DTM_PARTIC_MISSING 103005149
#define DTM_TMTX_ROLLBACK_BRANCHES_MULTIPASS 103005150
#define DTM_TXTHREAD_BAD_EVENT 103005151
#define DTM_TXTHREAD_BAD_STATE 103005152
#define DTM_LOGIC_ERROR 103005153
#define DTM_TMTX_TX_ROLLBACKONCOMMIT 103005154
#define DTM_TMTX_PREPARE_BRANCHES_MULTIPASS 103005155
#define DTM_TMTX_COMMIT_BRANCHES_MULTIPASS 103005156
#define DTM_TM_SHUTDOWN_ABORT_TXNS 103005157
#define DTM_TM_SHUTDOWNP1WAIT_RUNNING 103005158
#define DTM_TM_SHUTDOWNP1WAIT_NONLEAD_ERR 103005159
#define DTM_HBASE_BEGIN_TXN_ERROR 103005160
#define DTM_HBASE_ATTACHTHREAD_FAILED 103005161
#define DTM_HBASE_DETACHTHREAD_FAILED 103005162
#define DTM_DUPLICATE_SHUTDOWN_NOTICE 103005163
#define DTM_DUPLICATE_CALL_MSG_DESTRUCTOR 103005164
#define TM_HBASE_COMMIT_CONFLICT 103005165
#define TM_HBASE_PREPARE_FAIL 103005166
#define DTM_TM_AUD_INVALID_TRANS_STATE 103005201
#define DTM_RM_NO_MATCH 103005202
#define DTM_SYNC_INVALID_DATA 103005203
#define DTM_TM_TEST_FORCE_ABORT 103005204
#define DTM_SYNC_INVALID_PID 103005205
#define DTM_SYNC_INVALID_TRANSID 103005206
#define DTM_TM_NODE_OUTSIDE_RANGE 103005207
#define DTM_TM_TEST_REC_START_SYNC 103005208
#define DTM_TM_UNKNOWN_SYNC_TYPE 103005209
#define DTM_TM_START_NODE_RECOVERY 103005210
#define DTM_TM_END_NODE_RECOVERY 103005211
#define DTM_TM_INVALID_TRANSACTION 103005212
#define DTM_TM_REGISTRY_SET_ERROR 103005213
#define DTM_TM_EXCEEDED_SYNC_ABORT_TRIES 103005214
#define DTM_INVALID_PROC_MON_MSG 103005215
#define DTM_TM_READ_MSG_FAIL 103005216
#define DTM_TM_MSG_VERSION_INVALID 103005217
#define DTM_LIB_TRANS_INVALID_ID 103005218
#define DTM_LIB_INVALID_TRANS 103005219
#define DTM_SEA_SOFT_FAULT 103005220
#define DTM_TM_INFO_INVALID_TRANSID 103005221
#define DTM_TM_INFO_LINK_MSG_FAIL 103005222
#define DTM_TM_INFO_NO_LTAG 103005223
#define DTM_CP_TO_TM_FAILED 103005224
#define DTM_LIBTXN_INVALID_TXN 103005225
#define DTM_LIBTXN_INVALID_IP_TXN 103005226
#define DTM_TSEBRANCH_RMIDNOTFOUND 103005227
#define DTM_RECOV_INVALID_LEAD_TM 103005228
#define DTM_RECOV_NOT_FORGOT_STATE 103005229
#define DTM_RECOV_INVALID_OLD_CMFG_STATE 103005230
#define DTM_RECOV_INVALID_OLD_ABRT_STATE 103005231
#define DTM_SHTDWN_INVALID_TM_INFO 103005232
#define DTM_SHTDWN_INVALID_LINKTAG 103005233
#define DTM_SHTDWN_NO_MON 103005234
#define DTM_THRD_INVALID_MSG_TYPE 103005235
#define DTM_THRD_INVALID_THREAD 103005236
#define DTM_TRANS_ENDER 103005237
#define DTM_TRANS_BAD_TRANS_SEQ 103005238
#define DTM_SEAPI_INVALID_STRING_SIZE 103005239
#define DTM_SELIB_DEF_CONS 103005240
#define DTM_SELIB_TSE_MSG_TOO_LONG 103005241
#define DTM_XMAPI_INVALID_STRING_SIZE 103005242
#define DTM_XMAPI_INVALID_RM_OBJ 103005243
#define DTM_XATMLIB_INVALID_RM_MSG 103005244
#define DTM_XATMLIB_INVALID_RET_COUNT 103005245
#define DTM_XATMLIB_INVALID_PTR 103005246
#define DTM_TMTX_INVALID_MSGID 103005247
#define DTM_TMTX_PROG_ERROR 103005248
#define DTM_TMTX_GETRM_FAIL 103005249
#define DTM_TMTX_INVALID_BRANCH 103005250
#define DTM_TMTX_INVALID_RMID 103005251
#define DTM_TMTX_BRANCH_STATE_UNKNOWN 103005252
#define TM_TEST_XA_OK_ASSERT 103005253
#define TM_TMTX_BRCH_STRT_UKN_STATE 103005254
#define TM_TMTX_COMP_NOT_BEG 103005255
#define TM_TMTX_COMMIT_NO_OTHERS 103005256
#define TM_TMTX_PHS1_COMMIT 103005257
#define TM_TMTX_INVALID_STATE 103005258
#define TM_TMTX_RLL_BACK_NO_OTHERS 103005259
#define TM_TMTX_INVALID_FGR_STATE 103005260
#define TM_TMTX_INVALID_MSG_TYPE 103005261
#define DTM_TSE_DOOMEDTXN 103005262
#define DTM_TM_RESTARTED 103005263
#define DTM_RECOV_LOST_LEAD_TM 103005264
#define DTM_RECOV_RCV_NULL_TX_INFO 103005265
#define DTM_TM_INFO_CNTRL_PTRSP_FAIL 103005266
#define DTM_TM_MEM_ALLOCATION_FAILED 103005267
#define DTM_TMTX_INVALID_POINTER 103005268
#define DTM_AUDIT_FAILED_WRITE 103005269
#define DTM_TM_UKN_SYNC_TYPE 103005270
#define DTM_TM_LEADTM_SET 103005271
#define DTM_TMRESTART_ERROR 103005273
#define DTM_TMTIMER_BAD_EVENT 103005274
#define DTM_TMTIMER_BAD_EVENT2 103005275
#define DTM_GETTIME_FAIL 103005276
#define DTM_TMTIMER_FAILED 103005277
#define DTM_BAD_PARAM 103005278
#define DTM_LOGIC_ERROR_TX_OBJ 103005279
#define DTM_TMTX_TX_FORGET_BAD_RESP 103005280
#define DTM_TMTX_INVALID_STATE_CURR_REQ 103005281
#define DTM_BCAST_MSG_TO_DTM_FAILED 103005282
#define DTM_TMTX_TX_HUNGCOMMITTED 103005283
#define DTM_TMTX_TX_HUNGCOMMITTED_ERROR 103005284
#define DTM_TMTX_TX_HUNGABORTED 103005285
#define DTM_TMTX_TX_HUNGABORTED_ERROR 103005286
#define DTM_XATM_COMPLETEALL_FAILED_UNKNOWN_TAG 103005287
#define DTM_XATM_SENDTORM_FAILED_LARGE_BUF 103005288
#define DTM_TXTHREAD_BAD_EVENT_RELEASE 103005289
#define DTM_TMTIMER_BAD_EVENT_NO_TH 103005290
#define DTM_TMTIMER_UNEXP_EVENT 103005291
#define DTM_SONAR_INSTANTIATE_ERROR 103005292
#define DTM_LOGIC_ERROR_THR_MISMATCH 103005293
#define DTM_LOGIC_ERROR_NO_ENTRY 103005294
#define DTM_TMTX_PROG_ERROR_RM_MSG 103005295
#define DTM_TM_INFO_CNTRL_PT_ISOLATED 103005296
#define DTM_RECOV_TOO_MANY_TXNS 103005297
#define DTM_TMTX_TX_HUNG_ERROR 103005298
#define DTM_LIBTXN_INVALID_ENLISTED_TXN 103005299
#define DTM_RECOV_TXN_FOUND_RECOV_IGNORED 103005300
#define DTM_TM_INFO_MSGBRK_FAIL2 103005301
#define DTM_TM_INFO_MSGBRK_FAIL3 103005302
#define DTM_TM_INFO_MSGBRK_FAIL4 103005303
#define DTM_TMTX_GETRM_FAIL2 103005304
#define DTM_PROCDEATH_ANY 103005305
#define DTM_ROLLOVER_LINK_TO_DTM_FAILED 103005306
#define DTM_ROLLOVER_OPENPROC_FAILED 103005307
#define DTM_BEGINTX_PROC_NOTEXIST 103005308
#define DTM_ROLLOVER_CP_ERROR 103005309
#define DTM_TMRESTARTED 103005310
#define DTM_TM_JNI_ERROR  103005311

/**********************************************/
/**********************************************/


/*********** TSE ***********/
#define TSE_AT_THRESHOLD_EXCEEDED 104002000
#define TSE_FIRST 104005008
#define TSE_BAD_UNDO_AREA 104005008
#define TSE_BAD_VLAB_SECT 104005010
#define TSE_BAD_VOL_NAME 104005011
#define TSE_BADDOWN_TMF 104005012
#define TSE_INTERNAL_ERR 104005022
#define TSE_MEMORY_PRESSURE 104005028
#define TSE_NO_MEMORY 104005029
#define TSE_UNABLE_READ_DIR_LABEL 104005036
#define TSE_UNABLE_WRITE_VLAB 104005037
#define TSE_BADDOWN_ADP 104005040
#define TSE_BAD_LABEL 104005048
#define TSE_BADDOWN_VOLUME 104005051
#define TSE_BAD_FILE_BLOCK 104005052
#define TSE_DATA_LOSS 104005053
#define TSE_WRITE_FAILURE 104005054
#define TSE_STR_FILE_OPEN_UNSTR 104005055
#define TSE_FL_AUDITED_FLAG_CHG 104005056
#define TSE_TESTPT 104005057
#define TSE_CACHE_INIT_FAILURE 104005059
#define TSE_DEFAULT_CACHE_SIZES 104005060
#define TSE_UNEXPECTED_MSG_ERROR 104005063
#define TSE_DV_ADPINFO_FROMTMFLIB 104005064
#define TSE_UNKNOWN_LABEL 104005065
#define TSE_REFRESH 104005066
#define TSE_NO_CBPOOL_SPACE 104005067
#define TSE_CBPOOL_USAGE 104005068
#define TSE_FCB_TABLE_USE 104005069
#define TSE_OCB_TABLE_USE 104005070
#define TSE_NO_LKTABLE_SPACE 104005071
#define TSE_LK_TABLE_USE 104005072
#define TSE_LAB_TABLE_USE 104005073
#define TSE_LNB_TABLE_USE 104005074
#define TSE_NO_ACPOOL_SPACE 104005075
#define TSE_ACPOOL_USAGE 104005076
#define TSE_DPSHARE_SIZE 104005077
#define TSE_NO_VIRTUAL_SPACE 104005078
#define TSE_NO_WORKPOOL_SPACE 104005079
#define TSE_WORKPOOL_USAGE 104005080
#define TSE_TERMINATED 104005082
#define TSE_BROTHER_TERMINATED 104005083
#define TSE_PROCESS_CREATE_TP 104005084
#define TSE_PROCESS_CREATE_BK 104005085
#define TSE_RESIZE_FAILED 104005086
#define TSE_ALLOCATE_FAILED 104005087
#define TSE_CURRENT_STATE 104005088
#define TSE_SOFTDOWN_RESET 104005089
#define TSE_OPER_ABORTED_DUMP 104005090
#define TSE_TSE_ABORTED_DUMP 104005091
#define TSE_DUMP_TIMEOUT 104005092
#define TSE_SOFTDOWN_STATE 104005093
#define TSE_NO_FLAT_SEGMENTS 104005094
#define TSE_CHECKPOINT_TIMEOUT 104005097
#define TSE_MAX_LOCK_LIMIT 104005160
#define TSE_REVIVE_STARTED 104005161
#define TSE_REVIVE_SUCCESS 104005162
#define TSE_REVIVE_FAIL    104005163
#define TSE_REVIVE_BAD_FILE 104005164
#define TSE_TFDS_DUMP_STARTED 104005165
#define TSE_TFDS_DUMP_COMPLETED 104005166
#define TSE_GIVE_OWNERSHIP_FAILED 104005167
#define TSE_FULL_HALT_REDUCTION 104005168
#define TSE_PER_ACTION_CONTINUE 104005169
#define TSE_MONITOR_CANNOT_MOUNT_DEVICE 104005170
#define TSE_CPU_STARTED 104005171
#define TSE_IGNORE_SECURITY 104005172
#define TSE_BACKOUT_FAILED 104005173
#define TSE_RETRY_SEMAPHORE 104005174
#define TSE_USER_TERMINATE 104005175
#define TSE_FILES_OPEN 104005176
#define TSE_TX_ABORTED 104005177
#define TSE_TX_HUNG 104005178
#define TSE_NO_PARENT_DIR 104005179
#define TSE_CANNOT_CREATE_VOL 104005180
#define TSE_CANNOT_ONLINE_VOL 104005181
#define TSE_VOLUME_NOT_NTFS 104005182
#define TSE_CANNOT_OFFLINE_VOL 104005183
#define TSE_SHUTDOWN 104005184
#define TSE_COMMAND_LINE_ERROR 104005185
#define TSE_LOG_PER 104005186
#define TSE_NONFATAL_ERROR 104005187
#define TSE_WAIT_OPER_TERMINATE 104005188
#define TSE_WAIT_OPER_TERMINATE_REASON 104005189
#define TSE_LOG_TRACE 104005190
#define TSE_UNKNOWN_NT_ERROR 104005191
#define TSE_STARTUP 104005192
#define TSE_MEASURE_ERROR 104005193
#define TSE_PROCESS_STOP 104005194
#define TSE_VOLUME_UP 104005195
#define TSE_VOLUME_DOWN 104005196
#define TSE_SESSION_FIRST_RETRY 104005197
#define TSE_SESSION_NEXT_RETRY 104005198
#define TSE_MALLOC_ERROR 104005199
#define TSE_BACKOUT_CREATION_ERROR 104005200
#define TSE_VOLRCOV_CREATION_ERROR 104005201
#define TSE_BEGINTX_DISABLED 104005202
#define TSE_BEGINTX_ENABLED 104005203
#define TSE_NOUNDO_TX_TRIGGERED 104005204
#define TSE_TX_THRESHOLD_EXEEDED_2 104005205
#define TSE_TX_THRESHOLD_EXEEDED_3 104005206
#define TSE_COMP_HARDWARE_ACCESS_ERROR 104005207
#define TSE_COMP_HARDWARE_INIT_ERROR   104005208
#define TSE_COMP_HARDWARE_DISABLED     104005209
#define TSE_BACKOUT_TX_ABANDONED 104005210
#define TSE_LOCK_ESCALATION 104005211
#define TSE_BACKOUT_CREATION_LIMITATION_ERROR 104005212
#define TSE_VOLRCOV_CREATION_LIMITATION_ERROR 104005213
#define TSE_CPU_SWITCH_FAILURE_PROCESS_LOOPING 104005214
#define TSE_ACTT_THRESHOLD                     104005215
#define TSE_FCB_THRESHOLD                      104005216
#define TSE_LOCK_ERROR                         104005217
#define TSE_LOCK_CONFLICT                      104005218
#define TSE_LOCK_ESCALATION_ERROR              104005219

#define TSE_LAST 104005219

/**********************************************/
/**********************************************/


/*********** ASE ***********/


/**********************************************/
/**********************************************/


/*********** AMP ***********/

  #define AMP_AT_THRESHOLD_EXCEEDED             106000001
  #define AMP_AT_OUT_OF_SPACE                   106000002
  #define AMP_CANT_CREATE_BACKUP                106000003
  #define AMP_ERROR_OBTAINING_PHANDLE_RETRYING  106000004
  #define AMP_ERROR_OBTAINING_PHANDLE           106000005
  #define AMP_BUFFER_TOO_SMALL                  106000006
  #define AMP_BUFFER_TOO_SMALL_2                106000007
  #define AMP_FILE_CREATE_ERROR                 106000008
  #define AMP_FILE_CREATE_ERROR_2               106000009
  #define AMP_FILE_OPEN_ERROR                   106000010
  #define AMP_FILE_OPEN_ERROR_2                 106000011
  #define AMP_FILE_RENAME_ERROR                 106000012
  #define AMP_FILE_CLOSE_ERROR                  106000013
  #define AMP_REGISTRY_SET_ERROR                106000014
  #define AMP_REGISTRY_SET_ERROR_2              106000015
  #define AMP_REGISTRY_SET_ERROR_3              106000016
// #define AMP_REGISTRY_GET_ERROR              106000017
  #define AMP_LOGIC_INCONSISTENT                106000018
  #define AMP_LOGIC_INCONSISTENT_2              106000019
  #define AMP_UNEXPECTED_MESSAGE                106000020
  #define AMP_UNEXPECTED_MESSAGE_2              106000021
  #define AMP_UNEXPECTED_MESSAGE_3              106000022
  #define AMP_UNEXPECTED_MESSAGE_4              106000023
  #define AMP_UNEXPECTED_MESSAGE_5              106000024
  #define AMP_BULKREAD_ERROR                    106000025
  #define AMP_BULKWRITE_ERROR                   106000026
  #define AMP_CHECKPOINT_ERROR                  106000027
  #define AMP_DATA_CORRUPTION                   106000028
  #define AMP_READDATA_ERROR                    106000029
  #define AMP_READCTRL_ERROR                    106000030
  #define AMP_INVALID_MESSAGE_VERSION           106000031
  #define AMP_CONTROL_REQUEST_ERROR             106000032
  #define AMP_SETPARAM_ERROR                    106000033
  #define AMP_MSG_MON_ERROR_RETRY               106000034
  #define AMP_MSG_MON_ERROR_EXIT                106000035
  #define AMP_MSG_MON_ERROR_OPEN_BACKUP         106000036
  #define AMP_MSG_MON_ERROR_COMMANDLINE         106000037
  #define AMP_READ_AT_CONFIGURATION_ERROR       106000038
  #define AMP_ASE_STARTUP_MESSAGE_ERROR         106000039
  #define AMP_ASE_SHUTDOWN_MESSAGE_ERROR        106000040
  #define AMP_OPEN_CHECKPOINT_ERROR             106000041
  #define AMP_CLOSE_CHECKPOINT_ERROR            106000042
  #define AMP_AT_FILE_PREPARED                  106000043
  #define AMP_AMP_SHUTDOWN_MESSAGE_ERROR        106000044
  #define AMP_REDO_THRESHOLD_EXCEEDED           106000045
  #define AMP_AT_OUT_OF_SPACE_2                 106000046
  #define AMP_EMERGENCY_SHUTDOWN                106000047
  #define AMP_PREPARING_AT_FILE                 106000048




/**********************************************/
/**********************************************/

/*********  RECOVERY   ********/
#define RECOVERY_BACKOUT_NOT_LICENSED                        107000001
#define RECOVERY_CLIENT_NOT_LICENSED                         107000002
#define RECOVERY_BACKOUT_PFS_POOL_TOO_SMALL                  107000003
#define RECOVERY_CLIENT_PFS_POOL_TOO_SMALL                   107000004
#define RECOVERY_BACKOUT_VOLUME_ERROR                        107000005
#define RECOVERY_CLIENT_VOLUME_ERROR                         107000006
#define RECOVERY_BACKOUT_RESTART_EVENT                       107000007
#define RECOVERY_CLIENT_RESTART_EVENT                        107000008
#define RECOVERY_BACKOUT_FILE_ERROR                          107000009
#define RECOVERY_CLIENT_FILE_ERROR                           107000010
#define RECOVERY_BACKOUT_IGNORE_TRANSACTION_ERROR            107000011
#define RECOVERY_CLIENT_IGNORE_TRANSACTION_ERROR             107000012
#define RECOVERY_BACKOUT_RETRY_TRANSACTION_ERROR             107000013
#define RECOVERY_CLIENT_RETRY_TRANSACTION_ERROR              107000014
#define RECOVERY_BACKOUT_FAILURE                             107000015
#define RECOVERY_CLIENT_FAILURE                              107000016
#define RECOVERY_AUDIT_EXCEPTION                             107000017
#define RECOVERY_AUDIT_EXCEPTION_2                           107000018
#define RECOVERY_BACKOUT_READ_AUDIT_EXCEPTION                107000019
#define RECOVERY_CLIENT_READ_AUDIT_EXCEPTION                 107000020
#define RECOVERY_VRECOVERY_PFS_POOL_TOO_SMALL                107000021
#define RECOVERY_VRECOVERY_CLIENT_PFS_POOL_TOO_SMALL         107000022
#define RECOVERY_VRECOVERY_NOT_LICENSED                      107000023
#define RECOVERY_VRECOVERY_CLIENT_NOT_LICENSED               107000024
#define RECOVERY_VRECOVERY_VOLUME_ERROR                      107000025
#define RECOVERY_VRECOVERY_CLIENT_VOLUME_ERROR               107000026
#define RECOVERY_VRECOVERY_RESTART                           107000027
#define RECOVERY_VRECOVERY_CLIENT_RESTART                    107000028
#define RECOVERY_VRECOVERY_FILE_ERROR_POSITION               107000029
#define RECOVERY_VRECOVERY_CLIENT_FILE_ERROR_POSITION        107000030
#define RECOVERY_VRECOVERY_FILE_ERROR                        107000031
#define RECOVERY_VRECOVERY_CLIENT_FILE_ERROR                 107000032
#define RECOVERY_VRECOVERY_FAILURE                           107000033
#define RECOVERY_VRECOVERY_CLIENT_FAILURE                    107000034
#define RECOVERY_VRECOVERY_FAILURE_ENDFILEH_POSITION         107000035
#define RECOVERY_VRECOVERY_CLIENT_FAILURE_ENDFILEH_POSITION  107000036
#define RECOVERY_VRECOVERY_FAILURE_ENDFILEH                  107000037
#define RECOVERY_VRECOVERY_CLIENT_FAILURE_ENDFILEH           107000038
#define RECOVERY_FRECOVERY_FAILURE                           107000039
#define RECOVERY_ERROR_OBTAINING_PHANDLE_RETRYING            107000040
#define RECOVERY_ERROR_OBTAINING_PHANDLE                     107000041
#define RECOVERY_BACKOUT_CB_POOL_TOO_SMALL                   107000042
#define RECOVERY_CLIENT_CB_POOL_TOO_SMALL                    107000043


/*********** SQL/MX ***********/


/**********************************************/
/**********************************************/

/*********** MX/CS ***********/


/*********** lunmgr **********/
#define LUNMGR_MOUNT_EVENT              110001001
#define LUNMGR_MOUNT_ERROR_NODEV        110001002
#define LUNMGR_MOUNT_ERROR_DEVREM       110001003
#define LUNMGR_MOUNT_ERROR_GEN          110001004
#define LUNMGR_UNMOUNT_EVENT            110002001
#define LUNMGR_UNMOUNT_ERROR            110002002
#define LUNMGR_PATH_DOWN_EVENT          110003001
#define LUNMGR_DEVICE_REMOVAL_EVENT     110003002
#define LUNMGR_DEVICE_REMOVAL_ERROR     110003003
#define LUNMGR_DEVICE_CONNECTION_EVENT  110004001
#define LUNMGR_DEVICE_NOT_FOUND_ERROR   110005001


/*********** SECURITY **********/
/* Security Component = 11
 * Module: UAMS = 1
 */
#define SECURITY_EVENT_FILE_SYSTEM_STARTUP      111001001
#define SECURITY_EVENT_FILE_SYSTEM_IO           111001002
#define SECURITY_EVENT_INVALID_REQUEST_TYPE     111001003
#define SECURITY_EVENT_UAMS_ABNORMAL_EXIT       111001004
#define SECURITY_EVENT_INVALID_REQUEST_PARM     111001005
#define SECURITY_EVENT_PAM_ERROR                111001006
#define SECURITY_EVENT_SQL_ERROR                111001007

/*********** STFS **********/
/* STFS Component = 12 */

#define STFS_EVENT_STARTUP                      112001001
#define STFS_EVENT_SHUTDOWN                     112001002
#define STFS_EVENT_ERROR                        112001003
#define STFS_EVENT_NODE_DISKS_THRESHOLD         112001004
#define STFS_EVENT_NODE_DISKS_FULL              112001005
#define STFS_EVENT_CLUSTER_DISKS_THRESHOLD      112001006
#define STFS_EVENT_CLUSTER_DISKS_FULL           112001007

#define STFS_EVENT_SOFTWARE_FAILURE             112002001

/************************************************/
/*  Seapilot or Manageability Component = 13-14 */
/************************************************/

// Seapilot uses component numbers 13 and 14.  Within this range,
//    Seapilot subcomponents are currently:
//
// 113003XXX - Wrapper Library
// 114002XXX - mal library
// 114004XXX - protobuf serialization library
// 114006XXX - text protocol adapter
//
// There are also a set of reserved values:
// 113001000 - 113003050  Reserved
// 113005000 - 113005999  Reserved
// 114000000 - 114015999  Reserved
//

enum ErrorCode {
        SP_UNKNOWN                             = 113003000,
        SP_CONNECTION_CLOSED                   = 113003001,
        SP_CONFIG_NOT_AVAILABLE                = 113003002,
        SP_SEND_FAILED                         = 113003003,
        SP_SYNC_FAILED                         = 113003004,
        SP_INVALID_ROUTING_KEY                 = 113003005,
        SP_BAD_PARAM                           = 113003006,
        SP_EMPTRY_STRING                       = 113003007,
        SP_NOT_FOUND                           = 113003008,
        SP_CANNOT_GET_HOSTNAME                 = 113003009,
        SP_DISABLED                            = 113003010,
        SP_ROUTE_FAILED                        = 113003011,
        SP_FORMAT_INVALID                      = 113003012,
        SP_NOT_IMPLEMENTED                     = 113003013,
        SP_ALREADY_CONNECTED                   = 113003014,
        SP_BACKUP_ERROR                        = 113003022,
        SP_AUTHENTICATION_FAILED               = 113003034,
        SP_PROTO_IMPORT_FAIL                   = 113003500,
        SP_PROTO_MSG_NOT_INIT                  = 113003501,
        SP_PROTO_INVALID_FIELDNAME             = 113003502,
        SP_PROTO_INVALID_FIELDINDEX            = 113003503,
        SP_PROTO_INVALID_FIELDTYPE             = 113003504,
        SP_PROTO_REQUIRED_FIELD_EMPTY          = 113003505,
        SP_PROTO_INVALID_REPEATED_NUMBER       = 113003506,
        MAL_ERR_GETTING_ERRCODE                = 114002000,
        MAL_ARRAY_OUT_OF_BOUNDS                = 114002001,
        MAL_ASSERTION_FAILURE                  = 114002002,
        PS_FILE_OPEN_ERROR                     = 114004000,
        PS_FILE_CLOSE_ERROR                    = 114004001,
        PS_FILE_GET_POSITION_ERROR             = 114004002,
        PS_FILE_SEEK_ERROR                     = 114004003,
        PS_FILE_READ_ERROR                     = 114004004,
        PS_FILE_WRITE_ERROR                    = 114004005,
        PS_SYNTAX_PARSE_ERROR                  = 114004006,
        PS_TEXT_FORMAT_PRINT_ERROR             = 114004007,
        PS_TEXT_FORMAT_PARSE_WARNING           = 114004008,
        PS_TEXT_FORMAT_PARSE_ERROR             = 114004009,
        PS_OBJECT_CREATION_FAILURE             = 114004010,
        PS_PROTO_MESSAGE_DESCRIPTOR_FIND_ERROR = 114004011,

        TPA_ROUTINGKEY_CONVERSION_ERROR        = 114006000,
        TPA_ROUTINGKEY_MISSING_COMPONENT_ERROR = 114006001,
        TPA_ROUTINGKEY_PROTOCOL_ERROR          = 114006002,
        TPA_ROUTINGKEY_MISSING_PUBNAME_ERROR   = 114006003,
        TPA_MESSAGE_READ_ERROR                 = 114006004,
        TPA_MESSAGE_TRAILER_ERROR              = 114006005,
        TPA_MESSAGE_EXTRANEOUS_TEXT_ERROR      = 114006006,
        TPA_PUBNAME_IMPORT_ERROR               = 114006007,
        TPA_UNSUPPORTED_FIELD_ERROR            = 114006008,
        TPA_FIELD_ASSIGN_NULL_ERROR            = 114006009,
        TPA_MESSAGE_DESERIALIZE_ERROR          = 114006010,
        TPA_LISTENINGFIFO_ERROR                = 114006011,
        TPA_REPEATED_FIELD_MISSING_PART_ERROR  = 114006012,
        TPA_FIELD_NOT_FOUND_ERROR              = 114006013,
        TPA_AMQP_CONNECT_ERROR                 = 114006015,
        TPA_INVALID_REPEATED_NUMBER            = 114006016,
        TPA_REMOVE_FILE_ERROR                  = 114006017,
        TPA_PROTO_IMPORT_FAIL                  = 114006018,
        TPA_FIFO_CREATE_ERROR                  = 114006011,
        TPA_FIFO_OPEN_ERROR                    = 114006012,
        PTPA_PROTOBUF_ERROR                    = 114015001,
        PTPA_QPID_ERROR                        = 114015002,
        PTPA_XML_PARSING_EXCEPTION             = 114015003,

        SP_LAST                                = 114999999
     };

// IODRV EVENT NUMBER

#define IODRV_DEVICE_INIT         117000001
#define IODRV_DEVICE_STATE_CHANGE 117000002

/*********** DB SECURITY COMPONENT **********/
/* DB Security Component = 19
 * Module:
 */
enum DB_SECURITY_EVENTID
{
    // General DBS Errors
    DBS_GENERIC_ERROR                           = 119001001,
    DBS_NO_MEMORY_ERROR                         = 119001002,
    DBS_CANT_REFRESH_CONFIGURATION              = 119001003,

    // Authentication Errors
    DBS_NO_LDAP_SEARCH_CONNECTION               = 119002001,
    DBS_LDAP_SEARCH_ERROR                       = 119002002,
    DBS_NO_LDAP_AUTH_CONNECTION                 = 119002003,
    DBS_UNKNOWN_AUTH_STATUS_ERROR               = 119002004,
    DBS_AUTH_RETRY_SEARCH                       = 119002005,
    DBS_AUTH_RETRY_BIND                         = 119002006,
    DBS_AUTH_CONFIG                             = 119002007,

    // dbaudit Errors
    DBS_AMQP_ERROR                              = 119003001,

    // dbaudit info
    DBS_AUTHENTICATION_ATTEMPT                  = 119004001,

    // SQL Errors
    DBS_SQL_ERROR                               = 119005001

};

#endif  // EVL_SQLOG_EVENTNUM_H


